==> dspy_async-2.6.12.tar.gz: dspy_async-2.6.12/dspy/utils/callback.py <==
import functools
import inspect
import logging
import uuid
from contextvars import ContextVar
from typing import Any, Callable, Dict, Optional

import dspy
from dspy.dsp.utils import Settings

ACTIVE_CALL_ID = ContextVar("active_call_id", default=None)

logger = logging.getLogger(__name__)


class BaseCallback:
    """A base class for defining callback handlers for DSPy components.

    To use a callback, subclass this class and implement the desired handlers. Each handler
    will be called at the appropriate time before/after the execution of the corresponding component.  For example, if
    you want to print a message before and after an LM is called, implement `the on_llm_start` and `on_lm_end` handler.
    Users can set the callback globally using `dspy.settings.configure` or locally by passing it to the component
    constructor.


    Example 1: Set a global callback using `dspy.settings.configure`.

    ```
    import dspy
    from dspy.utils.callback import BaseCallback

    class LoggingCallback(BaseCallback):

        def on_lm_start(self, call_id, instance, inputs):
            print(f"LM is called with inputs: {inputs}")

        def on_lm_end(self, call_id, outputs, exception):
            print(f"LM is finished with outputs: {outputs}")

    dspy.settings.configure(
        callbacks=[LoggingCallback()]
    )

    cot = dspy.ChainOfThought("question -> answer")
    cot(question="What is the meaning of life?")

    # > LM is called with inputs: {'question': 'What is the meaning of life?'}
    # > LM is finished with outputs: {'answer': '42'}
    ```

    Example 2: Set a local callback by passing it to the component constructor.

    ```
    lm_1 = dspy.LM("gpt-3.5-turbo", callbacks=[LoggingCallback()])
    lm_1(question="What is the meaning of life?")

    # > LM is called with inputs: {'question': 'What is the meaning of life?'}
    # > LM is finished with outputs: {'answer': '42'}

    lm_2 = dspy.LM("gpt-3.5-turbo")
    lm_2(question="What is the meaning of life?")
    # No logging here because only `lm_1` has the callback set.
    ```
    """

    async def on_module_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when forward() method of a module (subclass of dspy.Module) is called.

        Args:
            settings: The dspy settings context to use for the call
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Module instance.
            inputs: The inputs to the module's forward() method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_module_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Any],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after forward() method of a module (subclass of dspy.Module) is executed.

        Args:
            settings: The dspy settings context to use for the call
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            outputs: The outputs of the module's forward() method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_lm_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when __call__ method of dspy.LM instance is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The LM instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the LM's __call__ method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_lm_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after __call__ method of dspy.LM instance is executed.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the LM's __call__ method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_adapter_format_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when format() method of an adapter (subclass of dspy.Adapter) is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Adapter instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the Adapter's format() method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_adapter_format_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after format() method of an adapter (subclass of dspy.Adapter) is called..

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the Adapter's format() method. If the method is interrupted
                by an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_adapter_parse_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when parse() method of an adapter (subclass of dspy.Adapter) is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Adapter instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the Adapter's parse() method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_adapter_parse_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after parse() method of an adapter (subclass of dspy.Adapter) is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the Adapter's parse() method. If the method is interrupted
                by an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_tool_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when a tool is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Tool instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the Tool's __call__ method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_tool_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after a tool is executed.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the Tool's __call__ method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    def on_evaluate_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when evaluation is started.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Evaluate instance.
            settings: The dspy settings object
            inputs: The inputs to the Evaluate's __call__ method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    def on_evaluate_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Any],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after evaluation is executed.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings object
            outputs: The outputs of the Evaluate's __call__ method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass


def with_callbacks(fn):
    @functools.wraps(fn)
    async def wrapper(instance, settings, *args, **kwargs):
        # Combine global and local (per-instance) callbacks.
        callbacks = settings.get("callbacks", []) + getattr(instance, "callbacks", [])

        # If no callbacks are provided, just call the function
        if not callbacks:
            return await fn(instance, settings, *args, **kwargs)

        # Generate call ID as the unique identifier for the call, this is useful for instrumentation.
        call_id = uuid.uuid4().hex

        inputs = inspect.getcallargs(fn, instance, settings, *args, **kwargs)
        inputs.pop("self")  # Not logging self as input
        inputs.pop("settings")

        for callback in callbacks:
            try:
                await _get_on_start_handler(callback, instance, fn)(call_id=call_id, instance=instance, settings=settings, inputs=inputs)

            except Exception as e:
                logger.warning(f"Error when calling callback {callback}: {e}")

        results = None
        exception = None
        try:
            parent_call_id = ACTIVE_CALL_ID.get()
            # Active ID must be set right before the function is called, not before calling the callbacks.
            ACTIVE_CALL_ID.set(call_id)
            results = await fn(instance, settings, *args, **kwargs)
            return results
        except Exception as e:
            exception = e
            raise exception
        finally:
            # Execute the end handlers even if the function call raises an exception.
            ACTIVE_CALL_ID.set(parent_call_id)
            for callback in callbacks:
                try:
                    await _get_on_end_handler(callback, instance, fn)(
                        call_id=call_id,
                        settings=settings,
                        outputs=results,
                        exception=exception,
                    )
                except Exception as e:
                    logger.warning(
                        f"Error when applying callback {callback}'s end handler on function {fn.__name__}: {e}."
                    )

    return wrapper


def _get_on_start_handler(callback: BaseCallback, instance: Any, fn: Callable) -> Callable:
    """Selects the appropriate on_start handler of the callback based on the instance and function name."""
    if isinstance(instance, dspy.LM):
        return callback.on_lm_start
    elif isinstance(instance, dspy.Evaluate):
        return callback.on_evaluate_start

    if isinstance(instance, dspy.Adapter):
        if fn.__name__ == "format":
            return callback.on_adapter_format_start
        elif fn.__name__ == "parse":
            return callback.on_adapter_parse_start
        else:
            raise ValueError(f"Unsupported adapter method for using callback: {fn.__name__}.")

    if isinstance(instance, dspy.Tool):
        return callback.on_tool_start

    # We treat everything else as a module.
    return callback.on_module_start


def _get_on_end_handler(callback: BaseCallback, instance: Any, fn: Callable) -> Callable:
    """Selects the appropriate on_end handler of the callback based on the instance and function name."""
    if isinstance(instance, (dspy.LM)):
        return callback.on_lm_end
    elif isinstance(instance, dspy.Evaluate):
        return callback.on_evaluate_end

    if isinstance(instance, (dspy.Adapter)):
        if fn.__name__ == "format":
            return callback.on_adapter_format_end
        elif fn.__name__ == "parse":
            return callback.on_adapter_parse_end
        else:
            raise ValueError(f"Unsupported adapter method for using callback: {fn.__name__}.")

    if isinstance(instance, dspy.Tool):
        return callback.on_tool_end

    # We treat everything else as a module.
    return callback.on_module_end

==> dspy_async-2.6.17.tar.gz: dspy_async-2.6.17/dspy/utils/callback.py <==
import functools
import inspect
import logging
import uuid
from contextvars import ContextVar
from typing import Any, Callable, Dict, Optional

import dspy
from dspy.dsp.utils import Settings

ACTIVE_CALL_ID = ContextVar("active_call_id", default=None)

logger = logging.getLogger(__name__)


class BaseCallback:
    """A base class for defining callback handlers for DSPy components.

    To use a callback, subclass this class and implement the desired handlers. Each handler
    will be called at the appropriate time before/after the execution of the corresponding component.  For example, if
    you want to print a message before and after an LM is called, implement `the on_llm_start` and `on_lm_end` handler.
    Users can set the callback globally using `dspy.settings.configure` or locally by passing it to the component
    constructor.


    Example 1: Set a global callback using `dspy.settings.configure`.

    ```
    import dspy
    from dspy.utils.callback import BaseCallback

    class LoggingCallback(BaseCallback):

        def on_lm_start(self, call_id, instance, inputs):
            print(f"LM is called with inputs: {inputs}")

        def on_lm_end(self, call_id, outputs, exception):
            print(f"LM is finished with outputs: {outputs}")

    dspy.settings.configure(
        callbacks=[LoggingCallback()]
    )

    cot = dspy.ChainOfThought("question -> answer")
    cot(question="What is the meaning of life?")

    # > LM is called with inputs: {'question': 'What is the meaning of life?'}
    # > LM is finished with outputs: {'answer': '42'}
    ```

    Example 2: Set a local callback by passing it to the component constructor.

    ```
    lm_1 = dspy.LM("gpt-3.5-turbo", callbacks=[LoggingCallback()])
    lm_1(question="What is the meaning of life?")

    # > LM is called with inputs: {'question': 'What is the meaning of life?'}
    # > LM is finished with outputs: {'answer': '42'}

    lm_2 = dspy.LM("gpt-3.5-turbo")
    lm_2(question="What is the meaning of life?")
    # No logging here because only `lm_1` has the callback set.
    ```
    """

    async def on_module_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when forward() method of a module (subclass of dspy.Module) is called.

        Args:
            settings: The dspy settings context to use for the call
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Module instance.
            inputs: The inputs to the module's forward() method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_module_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Any],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after forward() method of a module (subclass of dspy.Module) is executed.

        Args:
            settings: The dspy settings context to use for the call
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            outputs: The outputs of the module's forward() method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_lm_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when __call__ method of dspy.LM instance is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The LM instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the LM's __call__ method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_lm_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after __call__ method of dspy.LM instance is executed.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the LM's __call__ method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_adapter_format_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when format() method of an adapter (subclass of dspy.Adapter) is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Adapter instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the Adapter's format() method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_adapter_format_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after format() method of an adapter (subclass of dspy.Adapter) is called..

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the Adapter's format() method. If the method is interrupted
                by an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_adapter_parse_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when parse() method of an adapter (subclass of dspy.Adapter) is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Adapter instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the Adapter's parse() method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_adapter_parse_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after parse() method of an adapter (subclass of dspy.Adapter) is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the Adapter's parse() method. If the method is interrupted
                by an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    async def on_tool_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when a tool is called.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Tool instance.
            settings: The dspy settings context to use for the call
            inputs: The inputs to the Tool's __call__ method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    async def on_tool_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Dict[str, Any]],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after a tool is executed.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings context to use for the call
            outputs: The outputs of the Tool's __call__ method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass

    def on_evaluate_start(
        self,
        call_id: str,
        instance: Any,
        settings: Settings,
        inputs: Dict[str, Any],
    ):
        """A handler triggered when evaluation is started.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            instance: The Evaluate instance.
            settings: The dspy settings object
            inputs: The inputs to the Evaluate's __call__ method. Each arguments is stored as
                a key-value pair in a dictionary.
        """
        pass

    def on_evaluate_end(
        self,
        call_id: str,
        settings: Settings,
        outputs: Optional[Any],
        exception: Optional[Exception] = None,
    ):
        """A handler triggered after evaluation is executed.

        Args:
            call_id: A unique identifier for the call. Can be used to connect start/end handlers.
            settings: The dspy settings object
            outputs: The outputs of the Evaluate's __call__ method. If the method is interrupted by
                an exception, this will be None.
            exception: If an exception is raised during the execution, it will be stored here.
        """
        pass


def with_callbacks(fn):
    @functools.wraps(fn)
    async def wrapper(instance, settings, *args, **kwargs):
        # Combine global and local (per-instance) callbacks.
        callbacks = settings.get("callbacks", []) + getattr(instance, "callbacks", [])

        # If no callbacks are provided, just call the function
        if not callbacks:
            return await fn(instance, settings, *args, **kwargs)

        # Generate call ID as the unique identifier for the call, this is useful for instrumentation.
        call_id = uuid.uuid4().hex

        inputs = inspect.getcallargs(fn, instance, settings, *args, **kwargs)

        # We don't include the instance information in the inpus
        if "self" in inputs:
            inputs.pop("self")
        elif "instance" in inputs:
            inputs.pop("instance")
        inputs.pop("settings", None)

        for callback in callbacks:
            try:
                await _get_on_start_handler(callback, instance, fn)(call_id=call_id, instance=instance, settings=settings, inputs=inputs)

            except Exception as e:
                logger.warning(f"Error when calling callback {callback}: {e}")

        results = None
        exception = None
        try:
            parent_call_id = ACTIVE_CALL_ID.get()
            # Active ID must be set right before the function is called, not before calling the callbacks.
            ACTIVE_CALL_ID.set(call_id)
            results = await fn(instance, settings, *args, **kwargs)
            return results
        except Exception as e:
            exception = e
            raise exception
        finally:
            # Execute the end handlers even if the function call raises an exception.
            ACTIVE_CALL_ID.set(parent_call_id)
            for callback in callbacks:
                try:
                    await _get_on_end_handler(callback, instance, fn)(
                        call_id=call_id,
                        settings=settings,
                        outputs=results,
                        exception=exception,
                    )
                except Exception as e:
                    logger.warning(
                        f"Error when applying callback {callback}'s end handler on function {fn.__name__}: {e}."
                    )

    return wrapper


def _get_on_start_handler(callback: BaseCallback, instance: Any, fn: Callable) -> Callable:
    """Selects the appropriate on_start handler of the callback based on the instance and function name."""
    if isinstance(instance, dspy.LM):
        return callback.on_lm_start
    elif isinstance(instance, dspy.Evaluate):
        return callback.on_evaluate_start

    if isinstance(instance, dspy.Adapter):
        if fn.__name__ == "format":
            return callback.on_adapter_format_start
        elif fn.__name__ == "parse":
            return callback.on_adapter_parse_start
        else:
            raise ValueError(f"Unsupported adapter method for using callback: {fn.__name__}.")

    if isinstance(instance, dspy.Tool):
        return callback.on_tool_start

    # We treat everything else as a module.
    return callback.on_module_start


def _get_on_end_handler(callback: BaseCallback, instance: Any, fn: Callable) -> Callable:
    """Selects the appropriate on_end handler of the callback based on the instance and function name."""
    if isinstance(instance, (dspy.LM)):
        return callback.on_lm_end
    elif isinstance(instance, dspy.Evaluate):
        return callback.on_evaluate_end

    if isinstance(instance, (dspy.Adapter)):
        if fn.__name__ == "format":
            return callback.on_adapter_format_end
        elif fn.__name__ == "parse":
            return callback.on_adapter_parse_end
        else:
            raise ValueError(f"Unsupported adapter method for using callback: {fn.__name__}.")

    if isinstance(instance, dspy.Tool):
        return callback.on_tool_end

    # We treat everything else as a module.
    return callback.on_module_end
